//
//	tags.cc: test the Tags classes
//
//	DESCRIPTION:
//		test the Tags classes by converting a number of frame counters
//		to proper time tags - the frames in the dummy ODF EMOS1 auxiliary
//		file are taken for taken for that purpose
//
//	CHANGE HISTORY:
//		UL, 1998-08-19: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Wed Aug 19 09:41:52 MET DST 1998
//

#include <iostream>
#include <cassert>
#include <cmath>
#include "errstr.h"
#include "TimeTags.h"
#include "OalServer.h"

int main()
{
	const Odf *odf = oalServer.newOdf("../odf");
	errstr<<*odf<<Sas::message(Msg::App, Msg::Verbose);

	//
	//	read the frame numbers in the dummy EMOS1 auxiliary file
	//
	const OdfSet *aux = odf->selectSet(Xmm::EMOS1, Xmm::AUXILIARY);
	assert(aux);

	DataSet *set = dataSetServer->open(aux->name(), DataSet::Read);
	Table *tab = set->table("AUXILIARY");

	ColumnData *cdframes = tab->column("FRAME")->data();
	int32 *frames = cdframes->int32Data();

	const FrameCounterVector *frameVec = oalServer.newFrameCounterVector(
											frames, tab->rows());

	ObtVector obtvec(frameVec->toObt());

	vector<double> metVec(obtvec.toMET());

	assert(metVec.size() == tab->rows());
	cout<<"Frame counters - OBT - tag - time:\n";
	cout.precision(16);
	for (unsigned n=0; n<metVec.size(); ++n) {
		STime t = Xmm::missionReferenceTime + metVec[n];
		cout<<frames[n]<<" "<<obtvec[n]<<" "<<metVec[n]<<" "
			<<t.toString()<<endl;
	}

	//
	//	check four counters
	//
	#define CHECK_FRAME(a, b, c, d) { \
		cout<<setprecision(12)<<fabs(obtvec[a]/b-1.)<<' '<<fabs(-metVec[a]/-c-1.)<<endl;\
		assert(fabs(obtvec[a]/b-1.) < 1.e-5 && fabs(-metVec[a]/-c-1.)< 1.e-5);\
		STime t = Xmm::missionReferenceTime + metVec[a]; \
		cout<<t.toString(1)<<' '<<d<<endl; \
		assert(t.toString(1) == d); \
	}
#if 0
	CHECK_FRAME(0, 2147483648.333328, -315619208.1772774,
				"1987-12-31T23:59:60.8");
	CHECK_FRAME(5, 2147516417.466648, -315586110.7526166,
				"1988-01-01T09:12:25.3");
	CHECK_FRAME(9, 2147516420.048888, -315586108.1445069,
				"1988-01-01T09:12:27.9");
	CHECK_FRAME(10, 2147516420.048928, -315586108.1444664,
				"1988-01-01T09:12:27.9");
#endif
	delete cdframes;
	dataSetServer->close(set);

	cout<<"\nAll tests successfully passed.\n";

	return 0;
}
