//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OalF90cxxGlue.cc	--- the Fortran90/C++ interface of the OAL
//
//	DESCRIPTION:
//		This file acts as the interface between the OAL API written in
//		Fortran90 and its implementation in C++. All platform-specifics
//		of the C++/F90 interface are centralized in this file.
//
//	CHANGE HISTORY:
//		UL, 1997-11-25: 0.0.1: - skeleton
//		UL, 1997-??-??: 1.0.0: - first version
//		UL, 1998-08-20: 2.0.0: - brought in-line with oal.f90 v2.0.0
//		UL, 1998-09-30: 2.1.0: - new inquiry function OAL_proposalInfo added
//		UL, 1998-10-15: 2.2.0: - added fields 'scheduledExposureLength'/
//								 'actualExposureLength' to 'ExposureInfoType'
//								 this is to compensate lack of availability
//								 of STime functionality in F90
//							   - added 'isSlew' field to 'OdfInfoType'
//							   - added inquiry function 'OAL_activeFilter'
//		UL, 1999-02-22: 2.3.0: - dropped support for fields
//								 'scheduledExposureLength'/
//								 'actualExposureLength'
//		UL, 1999-03-10: 2.4.0: - minor change in openOdf()
//		UL, 1999-03-17: 2.5.0: - in several routines: behave properly if
//								 input vector is empty
//							   - better error messages if input arguments
//								 are missing
//		UL, 1999-04-01: 2.6.0: - new functions: OAL_setState1/2/3 for
//								 defining state from explicit set of
//								 values/name of data set/reference to Block
//		UL, 1999-06-03: 2.7.0: - removed obsolete length arguments from
//							     vector-processing routines
//							   - added wrapper for ertToTimeTag
//		UL, 1999-06-11: 2.8.0: - getPosition() makes use of XmmOrbitSimulator
//		UL, 1999-08-18: 2.7.0: - made OAL_getPosition/getAttitude logical
//								 functions
//		UL, 1999-12-17: 2.8.0: - setState from block/set/name modified
//								 such that is makes use of the new
//								 functionality in OalServer
//		UL, 2000-05-31: 2.9.0: - added
//									OAL_hasAssociatedSet()
//									OAL_selectAssociatedSet()
//		UL, 2000-06-15: 2.10.0:- added
//									OAL_hasSet()
//		UL, 2000-09-26: 2.11.0:- removed obsolete OAL_activeFilter()
//							   - added OAL_getIPPVint()
//							   - made OAL_getIPPVstring() a function
//							   - added OAL_expandSetName()
//		UL, 2001-03-23: 2.12.0:- added OAL_getBadPixelList()
//							   - fixed bug in OAL_frameCounterToObt()/
//								 OAL_ObtToTimeTag() [did not work with
//								 allocated output vector pointers]
//		UL, 2001-08-10: 2.13.0:- added OAL_selectScopeAndExposure
//							   - added argument 'exposureId' to OAL_setState()
//							   - added arguments 'scheduled'/'exposureNrOdf'
//								 to OAL_getState()
//		UL, 2002-01-07: 2.14.0:- added OAL_addCommonAttributes()
//		RDS, 2003-12-03: 2.15.0:- added OAL_getRGSDiagImage()
//		RDS, 2003-12-12: 2.15.1:- removed OAL_getRGSDiagImage() - only C++
//
//	AUTHOR:
//		U. Lammers, Tue Nov 25 16:27:48 MET 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/oal/f90/OalF90cxxGlue.cc,v 1.51 2003/12/12 10:45:07 rsaxton Exp $
//
#include <cassert>
#include <cstdio>
#include <string>
#include <vector>

#include "OalServer.h"
#include "common.h"
#include "Handle.h"
#include "CalOalUtility.h"
#include "OalF90cxxGlue.h"
#include "Xmm.h"
#include "Tracer.h"
#include "errstr.h"

using namespace CalOalUtility;

namespace {

bool isValidPointer(const void *p)
{
	return p && p != reinterpret_cast<void *>(0xffffffff);
}

bool checkDope(const Dope1 *dp)
{
	if (!dp)
		errstr<<"vector not properly passed"
			  <<Sas::fatal("IllegalPointerValue");

	//
	//	the following ugly cast is necessary to detect uninitialized pointers
	//	from f90 - they get set to 0xffffffff (32 bit even on Dec with
	//	64 bit pointers!)
	//
	char *uninitp = reinterpret_cast<char *>(0xffffffff);

	return dp->addr != 0 && dp->addr != uninitp;
} // checkDope()

}

static int ___dum = ios::sync_with_stdio(1);

//
//	---( control routines )-------------------------------------------------
//
void CXXFUNC(oal_openodf, OAL_openOdf)(const char *directoryName,
									   Integer3 length)
{
	oalServer.newOdf(f90string(directoryName, length));
	SAS_DEBUG(cout<<"New ODF:\n"<<oalServer.odf()<<endl;)
} // OAL_openODF()


void CXXFUNC(oal_setstate1, OAL_setState1)(const Integer3 *instrument,
										   const Integer3 *dataMode,
										   const Integer3 *exposureNr,
										   const Logical *scheduled,
										   const Integer3 *ccdNr,
										   const Integer3 *nodeNr,
										   const Logical *attitudeFromAhf,
										   const char *instrExposureId,
										   Integer3 length)
{
    #define SETVAR(setfunc, argvp, cast)	\
		if (argvp)							\
			oalServer.setfunc(cast(*argvp))

	SETVAR(instrument, instrument, static_cast<Xmm::Instrument>);
	SETVAR(dataMode, dataMode, static_cast<Xmm::DataMode>);
	SETVAR(exposureIndex, exposureNr,);
	if (scheduled)
		oalServer.scheduled(*scheduled != 0);
	if (instrExposureId) {
		string id(f90string(instrExposureId, length));
		if (id.length() <= 4)
			oalServer.exposureId(id);
		else
			oalServer.instrExposureId(id);
	}
	SETVAR(ccd, ccdNr,);

	if (nodeNr)
		oalServer.node(Node(*nodeNr));
	if (attitudeFromAhf)
		oalServer.attitudeFromStr(*attitudeFromAhf != 0);

	#undef SETVAR
} // OAL_setState()


void CXXFUNC(oal_setstate2, OAL_setState2)
						(char *dataSetName, Integer3 length)
{
	if (!isValidPointer(dataSetName))
		errstr<<"No data set name given to OAL_setState"
			  <<Sas::error("InvalidInputArgument");

	oalServer<<f90string(dataSetName, length);
} // OAL_setState()


void CXXFUNC(oal_setstate3, OAL_setState3)(Handle<DataSet> *setP)
{
	DataSet *set = setP->addressOfObject();
	if (!isValidPointer(set))
		errstr<<"No valid data set pointer given to OAL_setState"
			  <<Sas::error("InvalidInputArgument");

	oalServer<<*set;
} // OAL_setState()


void CXXFUNC(oal_setstate4, OAL_setState4)(Handle<Block> *blockP)
{
	Block *block = blockP->addressOfObject();
	if (!isValidPointer(block))
		errstr<<"No valid block pointer given to OAL_setState"
			  <<Sas::error("InvalidInputArgument");

	oalServer<<*block;
} // OAL_setState()


void CXXFUNC(oal_setstatefromsetname, OAL_setStateFromSetName)
						(char *setName, Integer3 length)
{
	if (!isValidPointer(setName))
		errstr<<"No data set name given to OAL_setState"
			  <<Sas::error("InvalidInputArgument");

	oalServer.setStateFromSetName(f90string(setName, length));
} // OAL_setStateFromSetName()


void CXXFUNC(oal_getstate, OAL_getState)(Integer3 *instrument,
										 Integer3 *dataMode,           
										 Integer3 *exposureNr,
										 Logical *scheduled,
										 Integer3 *ccdNr,
										 Integer3 *nodeNr,
										 Logical *attitudeFromAhf,
										 Integer3 *exposureNrOdf)
{
	#define GETVAR(getfunc, argvalp)		\
		if (argvalp)						\
			*argvalp = oalServer.getfunc()

	GETVAR(instrument, instrument);
	GETVAR(dataMode, dataMode);
	GETVAR(exposureIndex, exposureNr);
	GETVAR(ccd, ccdNr);
	GETVAR(node, nodeNr);
	GETVAR(attitudeFromStr, attitudeFromAhf);
	GETVAR(scheduled, scheduled);
	GETVAR(exposureNumber, exposureNrOdf);

	#undef GETVAR
} // OAL_getState



//
// ---( ODF access )-------------------------------------------------------
//
Integer3 CXXFUNC(oal_selectscope, OAL_selectScope)(const Integer3 *instrument,
												   const Integer3 *dataMode)
{
	oalServer.instrument(static_cast<Xmm::Instrument>(*instrument));
	oalServer.dataMode(static_cast<Xmm::DataMode>(*dataMode));

	return oalServer.odf()->exposures(oalServer.instrument(),
									  oalServer.dataMode());
} // OAL_selectScope()



Integer3 CXXFUNC(oal_selectexposure, OAL_selectExposure)
								(const Integer3 *exposureNr)
{
	oalServer.exposureIndex(*exposureNr);

	return oalServer.odf()->sets(oalServer.instrument(),
								 oalServer.dataMode(),
								 oalServer.exposureIndex());
} // OAL_selectExposure()


Integer3 CXXFUNC(oal_selectscopeandexposure, OAL_selectScopeAndExposure)
				(const Integer3 *instrument,
				 const char *exposureId, Integer3 length)
{
	oalServer.instrument(static_cast<Xmm::Instrument>(*instrument));
	oalServer.exposureId(f90string(exposureId, length));

	return oalServer.odf()->sets(oalServer.instrument(),
								 oalServer.dataMode(),
								 oalServer.exposureIndex());
} // OAL_selectScopeAndExposure()


void CXXFUNC(oal_selectfile, OAL_selectFile)
								(const Integer3 *ccdNr,
								 char *fileName, Integer3 length)
{
	const OdfSet *odfSet = oalServer.odf()
			->selectSet(oalServer.instrument(), oalServer.dataMode(),
						oalServer.exposureNumber(), *ccdNr, Odf::EXP_ODF);
	string fullName(" ");
	if (odfSet) {
		fullName = odfSet->name();
		oalServer.node(Node(odfSet->nodeNr()));
	}

	stringf90(fullName, fileName, length);
} // OAL_selectFile()


Logical CXXFUNC(oal_hasset, OAL_hasSet)
				(const Integer3 *ccdNr)
{
	const OdfSet *odfSet = oalServer.odf()
			->selectSet(oalServer.instrument(), oalServer.dataMode(),
						oalServer.exposureNumber(), *ccdNr, Odf::EXP_ODF,
						false, true, false);

	return odfSet != 0;
} // OAL_hasSet()


void CXXFUNC(oal_selectset, OAL_selectSet)
								(char *fileName, Integer3 length,
								 const Integer3 *ccdNr)
{
	const OdfSet *odfSet = oalServer.odf()
			->selectSet(oalServer.instrument(), oalServer.dataMode(),
						oalServer.exposureNumber(), *ccdNr, Odf::EXP_ODF);
	string fullName(" ");
	if (odfSet) {
		fullName = odfSet->name();
		oalServer.node(Node(odfSet->nodeNr()));
	}

	stringf90(fullName, fileName, length);
} // OAL_selectSet()


Logical CXXFUNC(oal_hasassociatedset, OAL_hasAssociatedSet)
				(const Integer3 *dataMode)
{
	return oalServer.odf()->hasSet(static_cast<Xmm::DataMode>(*dataMode));
} // OAL_hasAssociatedSet()


void CXXFUNC(oal_selectassociatedset, OAL_selectAssociatedSet)
				 (char *fileName, Integer3 length,
				  const Integer3 *dataMode)
{
	const OdfSet *odfSet = oalServer.odf()
					->selectSet(static_cast<Xmm::DataMode>(*dataMode));
	string fullName(" ");
	if (odfSet)
		fullName = odfSet->name();
	stringf90(fullName, fileName, length);
} // OAL_selectAssociatedSet()


void CXXFUNC(oal_expandfilename, OAL_expandFileName)
				(const char *inFileName, char *outFileName,
				 Integer3 lengthInFileName, Integer3 lengthOutFileName)
{
	string s(f90string(inFileName, lengthInFileName));
	const OdfSet *odfSet = oalServer.odf()->selectSet(s);
	string fullName;
	if (odfSet)
		fullName = odfSet->name();
	stringf90(fullName, outFileName, lengthOutFileName);
} // OAL_expandFileName()


void CXXFUNC(oal_expandsetname, OAL_expandSetName)
				(char *setName, Integer3 length,
				 const char *inFileName, Integer3 lengthInFileName)
{
	string s(f90string(inFileName, lengthInFileName));
	const OdfSet *odfSet = oalServer.odf()->selectSet(s);
	string fullName;
	if (odfSet)
		fullName = odfSet->name();
	stringf90(fullName, setName, length);
} // OAL_expandSetName()


void CXXFUNC(oal_framecountertoobt, OAL_frameCounterToObt)
								(const Dope1 *frameCounters,
								 Dope1 *obts)
{
	assert(frameCounters);

	unsigned vecSize = frameCounters->dim[0].extent;

	const FrameCounterVector *frameCounterVector =
				oalServer.newFrameCounterVector(INTP(frameCounters->addr),
												vecSize);

	ObtVector obtVec(frameCounterVector->toObt());	//	do the conversion

	Double *obtVecOut = checkDope(obts)
							? DOUBLEP(obts->addr)	//	initialized pointer
							: new Double[vecSize];	//	uninitialized pointer
	for (unsigned i=0; i<vecSize; ++i)
		obtVecOut[i] = obtVec[i];

	//	setup the dope stuff
    vector<unsigned>	oneDim(1); oneDim[0] = vecSize;
	setupDope(*obts, obtVecOut, oneDim);

	//	and free up some memory
	delete frameCounterVector;
} // OAL_frameCounterToObt()


void CXXFUNC(oal_obttotimetag, OAL_obtToTimeTag)
								(const Dope1 *obts,
								 Dope1 *timeTags)
{
	assert(obts);
	assert(timeTags);

	unsigned vecSize = obts->dim[0].extent;

	//
	//	instantiate a dummy frame counter vector for getting an ObtVector
	//
	const FrameCounterVector *frameCounterVector =
									oalServer.newFrameCounterVector(0, 0);
	ObtVector obtVec(frameCounterVector);

	//	make enough room and copy input OBT values
	obtVec.reserve(vecSize);
	double *v = DOUBLEP(obts->addr);
	for (unsigned i=0; i<vecSize; obtVec.push_back(v[i++]));

	vector<double> utcVec(obtVec.toMET());
	double	*timeTagVec = checkDope(timeTags)
							? DOUBLEP(timeTags->addr)	//	initialized ptr
							: new Double[vecSize];		//	uninitialized ptr
	for (unsigned i=0; i<vecSize; ++i)
		timeTagVec[i] = utcVec[i];

	//	setup the dope stuff
    vector<unsigned>	oneDim(1); oneDim[0] = vecSize;
	setupDope(*timeTags, timeTagVec, oneDim);

	//	and free up some memory
	delete frameCounterVector;
} // OAL_ObtToTimeTag()


Double CXXFUNC(oal_erttotimetag, OAL_ertToTimeTag)
				(const char *ertString, Integer3 length)
{
	static const TimeCorrelator *tcorr = oalServer.timeCorrelator();

	string s(ertString, length);

	return tcorr->toMissionEllapsedTime(s);
} // OAL_ertToTimeTag()


Logical CXXFUNC(oal_getattitude, OAL_getAttitude)
								(const Double *timeTag,
								 SpacecraftAttitudeType *attitude)
{
	assert(timeTag);
	assert(attitude);

	attitude->ra = attitude->dec = attitude->apos = 0.;
	const AttitudeLocator *attitudeLocator = oalServer.attitudeLocator();
	SpacecraftAttitude att;
	try {
		att = attitudeLocator->attitude(*timeTag);
	} catch(...) {
		return static_cast<Logical>(false);
	}

	attitude->ra = att.ra();
	attitude->dec = att.dec();
	attitude->apos = att.apos();

	return static_cast<Logical>(true);
} // OAL_getAttitude()



Logical CXXFUNC(oal_getposition, OAL_getPosition)
								(const Double *timeTag,
								 CartesianVectorType *position,
								 CartesianVectorType *velocity,
								 Double *revolution)
{
	assert(timeTag);
	assert(position);

	position->x = position->y = position->z = 0.;
	if (velocity)
		velocity->x = velocity->y = velocity->z = 0.;
	const PositionLocator *positionLocator = oalServer.positionLocator();
	StateVector stateVector;
	try {
		stateVector = positionLocator->stateVector(*timeTag);
	} catch(...) {
		return static_cast<Logical>(false);
	}

	position->x = stateVector.position.x();
	position->y = stateVector.position.y();
	position->z = stateVector.position.z();

	if (velocity) {
		//
		//	get velocity vector as well
		//
		velocity->x = stateVector.velocity.x();
		velocity->y = stateVector.velocity.y();
		velocity->z = stateVector.velocity.z();
	}

	if (revolution)
		*revolution = stateVector.revolution;

	return static_cast<Logical>(true);
} // OAL_getPosition()


//
//	---( OAL_getBadPixelList )----------------------------------------------
//
Logical CXXFUNC(oal_getbadpixellist, OAL_getBadPixelList)
							(Dope1 *rawX, Dope1 *rawY,
							 Dope1 *type,
							 Dope1 *yextent, Dope1 *uplinked)
{
	if (!oalServer.odf()->hasBadPixelMap())
		return false;

	BadPixelMap map = oalServer.odf()->badPixelMap();
	unsigned nPix = map.size();

	//	allocate space for vectors
	Integer2 *rawxList		= new Integer2[nPix];
	Integer2 *rawyList		= new Integer2[nPix];
	Integer2 *yextentList	= new Integer2[nPix];
	Integer2 *typeList		= new Integer2[nPix];
	Logical  *uplinkedList	= new Logical[nPix];

	//	setup the dope stuff properly
	vector<unsigned>	oneDim(1); oneDim[0] = nPix;
	setupDope(*rawX, rawxList, oneDim);
	setupDope(*rawY, rawyList, oneDim);
	setupDope(*yextent, yextentList, oneDim);
	setupDope(*type, typeList, oneDim);
	setupDope(*uplinked, uplinkedList, oneDim);

	BadPixelMap::const_iterator i;
	unsigned n = 0;
	for (i=map.begin(); i!=map.end(); ++i) {
		rawxList[n]			= static_cast<Integer2>((*i).first.first);
		rawyList[n]			= static_cast<Integer2>((*i).first.second);
		yextentList[n]		= static_cast<Integer2>((*i).second.yextent());
		typeList[n]			= static_cast<Integer2>((*i).second.type());
		uplinkedList[n++]	= static_cast<Logical>((*i).second.uplinked());
	}

	return true;
} // OAL_getBadPixelList()

//
//	---( inquiry / utility functions )---------------------------------------
//
void CXXFUNC(oal_odfinfo, OAL_odfInfo)(OdfInfoType *odfInfo)
{
	assert(odfInfo);

	odfInfo->isSlew = oalServer.odf()->isSlew();
	odfInfo->revolutionNr = oalServer.odf()->orbitNr();
	odfInfo->nrFile = oalServer.odf()->sets();
	stringf90(oalServer.odf()->proposalId(), odfInfo->proposalId,
			  sizeof(odfInfo->proposalId));
	stringf90(oalServer.odf()->observationId(), odfInfo->observationId,
			  sizeof(odfInfo->observationId));
	stringf90(oalServer.odf()->observationStartTime().toString(),
			  odfInfo->obsStartTime, sizeof(odfInfo->obsStartTime));
	stringf90(oalServer.odf()->observationEndTime().toString(),
			  odfInfo->obsEndTime, sizeof(odfInfo->obsEndTime));
	stringf90(oalServer.odf()->directoryName(), odfInfo->directoryName,
			  sizeof(odfInfo->directoryName));
} // OAL_odfInfo



void CXXFUNC(oal_proposalinfo, OAL_proposalInfo)(ProposalInfoType
															*proposalInfo)
{
	assert(proposalInfo);

	ProposalInfo pi = oalServer.odf()->proposalInfo();

	stringf90(pi.GoName, proposalInfo->GoName, sizeof(proposalInfo->GoName));
	stringf90(pi.GoInstitute, proposalInfo->GoInstitute,
			  sizeof(proposalInfo->GoInstitute));
	stringf90(pi.GoAddress, proposalInfo->GoAddress,
			  sizeof(proposalInfo->GoAddress));
	stringf90(pi.GoEmail, proposalInfo->GoEmail,
			  sizeof(proposalInfo->GoEmail));
	stringf90(pi.AO, proposalInfo->AO, sizeof(proposalInfo->AO));
	stringf90(pi.scienceType, proposalInfo->scienceType,
			  sizeof(proposalInfo->scienceType));
	stringf90(pi.targetName, proposalInfo->targetName,
			  sizeof(proposalInfo->targetName));

	proposalInfo->targetRa = pi.targetRa;
	proposalInfo->targetDec = pi.targetDec;
	proposalInfo->obsDuration = pi.obsDuration;
	proposalInfo->boresightRa = pi.boresightRa;
	proposalInfo->boresightDec = pi.boresightDec;

	stringf90(pi.posAngleConstLo, proposalInfo->posAngleConstLo,
			  sizeof(proposalInfo->posAngleConstLo));
	stringf90(pi.posAngleConstHi, proposalInfo->posAngleConstHi,
			  sizeof(proposalInfo->posAngleConstHi));
	stringf90(pi.posAngleOrigRef, proposalInfo->posAngleOrigRef,
			  sizeof(proposalInfo->posAngleOrigRef));

	proposalInfo->emos1Prio = pi.emos1Prio;
	proposalInfo->emos2Prio = pi.emos2Prio;
	proposalInfo->epnPrio = pi.epnPrio;
	proposalInfo->rgs1Prio = pi.rgs1Prio;
	proposalInfo->rgs2Prio = pi.rgs2Prio;
	proposalInfo->omPrio = pi.omPrio;
} // OAL_proposalInfo



void CXXFUNC(oal_exposureinfo, OAL_exposureInfo)
				(ExposureInfoType *exposureInfo)
{
	assert(exposureInfo);

	const ExposureRecord *expRec = oalServer.odf()->
										exposure(oalServer.instrument(),
												 oalServer.dataMode(),
												 oalServer.exposureIndex());

	exposureInfo->number = expRec->counter();
	exposureInfo->scheduled = expRec->scheduled();
	exposureInfo->type = expRec->type() == ExposureRecord::SCIENCE ? 0 : 1;
	stringf90(expRec->scheduledStartTime().toString(),
			  exposureInfo->scheduledStartTime,
			  sizeof(exposureInfo->scheduledStartTime));
	stringf90(expRec->scheduledEndTime().toString(),
			  exposureInfo->scheduledEndTime,
			  sizeof(exposureInfo->scheduledEndTime));
	stringf90(expRec->actualStartTime().toString(),
			  exposureInfo->actualStartTime,
			  sizeof(exposureInfo->actualStartTime));
	stringf90(expRec->actualEndTime().toString(),
			  exposureInfo->actualEndTime,
			  sizeof(exposureInfo->actualEndTime));
} // OAL_exposureInfo()


Logical CXXFUNC(oal_hasippv, OAL_hasIPPV)
				(const char *keyword, Integer3 length)
{
	if (!keyword)
		errstr<<"Argument `keyword' is missing - cannot continue"
			  <<Sas::error("ArgumentMissing");

	const Exposure *exp = oalServer.odf()->
									exposure(oalServer.instrument(),
											 oalServer.dataMode(),
											 oalServer.exposureIndex());

	return exp->ippvCollection().hasIPPV(f90string(keyword, length));
} // OAL_hasIPPV()


Integer3 CXXFUNC(oal_getippvint, OAL_getIPPVint)
					(const char *keyword, Integer3 length)
{
	if (!keyword)
		errstr<<"Argument `keyword' is missing - cannot continue"
			  <<Sas::error("ArgumentMissing");

	const Exposure *exp = oalServer.odf()->
									exposure(oalServer.instrument(),
											 oalServer.dataMode(),
											 oalServer.exposureIndex());

	return
		exp->ippvCollection().getIPPV(f90string(keyword, length)).valueAsInt();
} // OAL_getIPPVint()


Double CXXFUNC(oal_getippvreal, OAL_getIPPVreal)
				(const char *keyword, Integer3 length)
{
	if (!keyword)
		errstr<<"Argument `keyword' is missing - cannot continue"
			  <<Sas::error("ArgumentMissing");

	const Exposure *exp = oalServer.odf()->
									exposure(oalServer.instrument(),
											 oalServer.dataMode(),
											 oalServer.exposureIndex());

	return exp->ippvCollection().
							getIPPV(f90string(keyword, length)).valueAsReal();
} // OAL_getIPPVreal()


void CXXFUNC(oal_getippvstring, OAL_getIPPVstring)
				(char *value, const Integer3 valueLength,
				 const char *keyword, Integer3 keyLength)
{
	if (!keyword)
		errstr<<"Argument `keyword' is missing - cannot continue"
			  <<Sas::error("ArgumentMissing");

	const Exposure *exp = oalServer.odf()->
									exposure(oalServer.instrument(),
											 oalServer.dataMode(),
											 oalServer.exposureIndex());

	
	stringf90(exp->ippvCollection().getIPPV(f90string(keyword, keyLength)).
												value(), value, valueLength);
} // OAL_getIPPVstring()


void CXXFUNC(oal_toattitudematrix, OAL_toAttitudeMatrix)
						(const SpacecraftAttitudeType *attitude,
						 Double *matrix)
{
	assert(attitude);
	assert(matrix);

	SpacecraftAttitude att(attitude->ra, attitude->dec, attitude->apos);
	AttitudeMatrix attmat = att.toAttitudeMatrix();

	const double *psrc = attmat.dataFlat();
	for (unsigned n=0; n<9; ++n)
		matrix[n] = psrc[n];
} // OAL_toAttitudeMatrix()


void CXXFUNC(oal_toeulerangles, OAL_toEulerAngles)
						(const Dope2 *m,
						 EulerAngs *eulerAngles)
{
	assert(m);
	assert(eulerAngles);

	double *v = DOUBLEP(m->addr);
	AttitudeMatrix att(v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7], v[8]);

	EulerAngles e = att.toEulerAngles(EulerAngles::Euler321);

	eulerAngles->phi = e.phi();
	eulerAngles->theta = e.theta();
	eulerAngles->psi = e.psi();
} // OAL_toEulerAngles()


void CXXFUNC(oal_addcommonattributes, OAL_addCommonAttributes)
				(Handle<DataSet> *setP)
{
	DataSet *set = setP->addressOfObject();

	*set<<oalServer;
} // OAL_addCommonAttributes()



//
//	release memory
//
#define RELDATA(a) if (a) { delete[] a->addr; a->addr = 0; }
#define RELEASEMEM(fname, cname, N)		\
	void CXXFUNC(fname, cname)			\
		(Dope##N *a1, Dope##N *a2, Dope##N *a3) \
		{ RELDATA(a1); RELDATA(a2); RELDATA(a3); }

RELEASEMEM(oal_releasememorylogicalonedim, OAL_releaseMemoryLogicalOneDim, 1)
RELEASEMEM(oal_releasememoryreal2onedim, OAL_releaseMemoryReal2OneDim, 1)
RELEASEMEM(oal_releasememoryint2onedim, OAL_releaseMemoryInt2OneDim, 1)

#undef RELEASEMEM
#undef RELDATA
