//-----------------------------------------------------------------------------
//
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//  OdfBrowseInfo.c --- Implentation of OdfInfo classes
//
//  DESCRIPTION:
//		This implements the classes in the OdfInfo namespace.
//
//  CHANGE HISTORY:
//		UL, 2001-07-02: 1.0.0: - initial version
//		JBa,2002-07-02: 1.0.1: - DEVELOPMENT-TRACK! Usage of sstream
//
//  AUTHOR:
//		U. Lammers, Mon Jul  2 12:00:02 MET DST 2001
//
//  REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//  $Header: /home/sasbuild/repositories/sasdev/sas/packages/oal/odfbrowse/OdfBrowseInfo.cc,v 1.9 2003/06/16 08:41:35 gvacanti Exp $
//
#include <sstream>
#include <algorithm>

#include "OdfInfo.h"
#include "Odf.h"
#include "OalServer.h"
#include "stringutil.h"
#include <sstream>

//
//	---( OdfSet/List )---------------------------------------------------------
//
class OalOdfSet : public OdfInfo::OdfDataSet, public OdfSet {
public:
	OalOdfSet(const OdfSet &odfSet) : OdfSet(odfSet) {}
	std::string name() const { return OdfSet::name(); }
	std::string dirname() const { return OdfSet::dirname(); }
	std::string basename() const { return OdfSet::basename(); }
	unsigned revolution() const { return OdfSet::revolutionNr(); }
	std::string proposalId()	const { return OdfSet::proposalId(); }
	Xmm::Instrument	instrument() const { return OdfSet::instrument(); }
	bool scheduled() const { return OdfSet::scheduled(); }
	unsigned exposure()	const { return OdfSet::exposureNr(); }
	unsigned short ccd() const { return OdfSet::ccdNr(); }
	unsigned node()	const { return OdfSet::nodeNr(); }
	Xmm::DataMode dataMode() const { return OdfSet::dataMode(); }
	OdfInfo::OdfDataSet::Type type() const {
						return (OdfInfo::OdfDataSet::Type)OdfSet::type(); }
	OdfInfo::OdfDataSet::Format format() const {
						return (OdfInfo::OdfDataSet::Format)OdfSet::format(); }
	bool isValid() const { return OdfSet::isValid(); }

	virtual ~OalOdfSet() {}
};

//
//	---( Exposure/List )-------------------------------------------------------
//
class OalExposure : public OdfInfo::ObservationExposure, public Exposure {
public:
	OalExposure(Xmm::Instrument instr,
				Xmm::DataMode dataMode,
				const Exposure &exp)
		: Exposure(exp), _instr(instr), _dataMode(dataMode) {}

	STime start() const { return actualStartTime(); }
	STime stop() const { return actualEndTime(); }

	OdfInfo::ObservationExposure::Type	type() const {
		return (OdfInfo::ObservationExposure::Type)Exposure::type(); }

	unsigned counter() const { return Exposure::counter(); }
	bool scheduled() const { return Exposure::scheduled(); }

	Xmm::Instrument	instrument() const { return _instr; }

	Xmm::DataMode dataMode() const { return _dataMode; }
	Xmm::Mode mode() const {
		if (ippvCollection().hasIPPV("MODE"))
			return
				Xmm::toMode(ippvCollection().getIPPV("MODE").valueAsString());

		return Xmm::MODE_NONE;
	}
	Filter filter()	const {
		if (ippvCollection().hasIPPV("FILTER"))
			return Xmm::toFilter(ippvCollection().getIPPV("FILTER").
															valueAsString());

		return Xmm::FILTER_NONE;
	}

	std::string id() const {
		stringstream s;
		s<<Xmm::toString(_instr, true)<<(scheduled() ? 'S' : 'U')
		 <<setw(3)<<setfill('0')<<counter();

		return s.str();
	}

	const OdfInfo::OdfDataSetList &associatedScienceSets() const {
		if (_scienceSets.empty()) {
			typedef const std::list<const OdfSet *> SetList;
			SetList &sl = sets();
			for (SetList::const_iterator i=sl.begin(); i!=sl.end();
				_scienceSets.push_back(new OalOdfSet(*(*i++))));
		}

		return _scienceSets;
	}

	const OdfInfo::OdfDataSetList &associatedAuxiliarySets() const {
		return _auxSets;
	}

	~OalExposure() {
		for (OdfInfo::OdfDataSetList::const_iterator i = _scienceSets.begin();
			 i != _scienceSets.end(); delete *i++);
		for (OdfInfo::OdfDataSetList::const_iterator i = _auxSets.begin();
			 i != _auxSets.end(); delete *i++);
	}

private:
	Xmm::Instrument			_instr;
	Xmm::DataMode			_dataMode;
	mutable OdfInfo::OdfDataSetList	_scienceSets, _auxSets;
}; 


struct laterExposure :
				public binary_function<const OdfInfo::ObservationExposure *,
									   const OdfInfo::ObservationExposure *,
									   bool> {
	bool operator()(const OdfInfo::ObservationExposure *exp1,
					const OdfInfo::ObservationExposure *exp2)
	const { return exp1->start() > exp2->start(); }
};


//
//	---( Proposal )------------------------------------------------------------
//
class OalProposalData : public OdfInfo::ProposalData, public ProposalInfo {
public:
	OalProposalData(const ProposalInfo &prop) : ProposalInfo(prop) {}
	std::string GoName() const { return ProposalInfo::GoName; }
	std::string GoInstitute() const { return ProposalInfo::GoInstitute; }
	std::string GoAddress() const { return ProposalInfo::GoAddress; }
	std::string GoCountry() const { return ProposalInfo::GOcountry(); }
	std::string GoEmail() const { return ProposalInfo::GoEmail; }
	std::string AO() const { return 	ProposalInfo::AO; }
	std::string scienceType() const { return ProposalInfo::scienceType; }

	std::string targetName()	const { return ProposalInfo::targetName; }
	double targetRa() const { return ProposalInfo::targetRa; }
	double targetDec() const { return ProposalInfo::targetDec; }
	double obsDuration() const { return ProposalInfo::obsDuration; }
	double boresightRa() const { return ProposalInfo::boresightRa; }
	double boresightDec() const { return ProposalInfo::boresightDec; }

	//	lower/upper/ref of position angle
	std::string posAngleConstLo() const { return ProposalInfo::posAngleConstLo; }
	std::string posAngleConstHi() const { return ProposalInfo::posAngleConstHi; }
	std::string posAngleOrigRef() const { return ProposalInfo::posAngleOrigRef; }

	//	instrument priorities
	short emos1Prio() const { return ProposalInfo::emos1Prio; }
	short emos2Prio() const { return ProposalInfo::emos2Prio; }
	short epnPrio()	const { return ProposalInfo::epnPrio; }
	short rgs1Prio() const { return ProposalInfo::rgs1Prio; }
	short rgs2Prio() const { return ProposalInfo::rgs1Prio; }
	short omPrio() const { return ProposalInfo::omPrio; }

	virtual	~OalProposalData() {}
};


//
//	---( Odf )-----------------------------------------------------------------
//
class OalOdf : public OdfInfo::Odf {
public:
	OalOdf() : _proposalData(oalServer.odf()->proposalInfo()) {
		//	construct map with exposures
		for (unsigned i=0; i<xmmPayLoad.instruments().size(); ++i) {
			XmmInstrument *ins = xmmPayLoad.instruments()[i];
			if (ins->id() == Xmm::ERM)
				continue;
			OdfInfo::ObservationExposureList sexplist;
			const std::vector<DataMode> &dm =
											ins->scienceEngineeringDataModes();
			for (unsigned n=0; n<dm.size(); ++n) {
				unsigned nexp = oalServer.odf()->exposures(ins->id(), dm[n]);
				for (unsigned i=1; i<=nexp; ++i) {
					OdfInfo::ObservationExposure *exp =
						new OalExposure(ins->id(), dm[n],
							*oalServer.odf()->exposure(ins->id(), dm[n], i));
						if ((ins->id() == Xmm::EMOS1 ||
							 ins->id() == Xmm::EMOS2)
							&& exp->associatedScienceSets().front()->ccd() >1
							&& exp->dataMode() == Xmm::IMAGING) {
							delete exp;
							continue;
						}
							
						typedef OdfInfo::ObservationExposureList::iterator
																		Iter;
						Iter pos =
								std::find_if(sexplist.begin(), sexplist.end(),
										 std::bind2nd(laterExposure(), exp));
						sexplist.insert(pos, exp);
				}
			}
			_expMap[ins->id()] = sexplist;
		}
	}

	std::string location() const {
		return oalServer.odf()->directoryName();
	}

	unsigned short revolution() const {
		return oalServer.odf()->orbitNr();
	}

	std::string proposalId() const {
		return oalServer.odf()->proposalId();
	}

	std::string observationId() const {
		return oalServer.odf()->observationId();
	}

	bool isSlew() const {
		return oalServer.odf()->isSlew();
	}

	const OdfInfo::ProposalData &proposalData() const {
		return _proposalData;
	}

	const OdfInfo::ObservationExposureList
									&observationExposures(Xmm::Instrument ins)
	const {
		return _expMap[ins];
	}

	STime start() const {
		return oalServer.odf()->observationStartTime();
	}

	STime stop() const {
		return oalServer.odf()->observationEndTime();
	}

	virtual ~OalOdf() {
		for (ExpMap::iterator i=_expMap.begin(); i != _expMap.end(); ++i)
			for (OdfInfo::ObservationExposureList::iterator j =
					i->second.begin(); j!=i->second.end(); delete *j++);
	}

private:
	OalProposalData	_proposalData;
	typedef map<Xmm::Instrument, OdfInfo::ObservationExposureList,
				less<Xmm::Instrument> >	ExpMap;
	mutable ExpMap			_expMap;
};


//
//	---( Server )-------------------------------------------------------------
//
class OalInfoServer : public OdfInfo::Server {
public:
	OalInfoServer() : _oalOdf(0) {}

	const OdfInfo::Odf *odf() const {
		if (!_oalOdf)
			return odf("");

		return _oalOdf;
	}

	const OdfInfo::Odf *odf(const std::string &location) const {
		oalServer.newOdf(location);
		delete _oalOdf;
		_oalOdf = new OalOdf;

		return _oalOdf;
	}

	virtual ~OalInfoServer() {
		delete _oalOdf;
	}

private:
	mutable OalOdf	*_oalOdf;
};

OdfInfo::Server *OdfInfo::server = new OalInfoServer;
