//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeCorrelator	--- user friendly front-end to ODF Time Correlation File
//
//	DESCRIPTION:
//		A class TimeCorrelator is defined which provides methods to perform
//		the following time-related conversions:
//
//			o OBTcoarse/OBTfine -> Mission Ellapsed Time tag
//			o ERT				-> Mission Ellapsed Time tag
//			o MET				-> On-Board Time tag
//
//		These conversions make use of the data in the Spacecraft Time
//		Correlation File which is part of every ODF.
//
//	CHANGE HISTORY:
//		UL, 1998-07-15: 1.0.0: - first version
//		UL, ...
//		UL, 2001-03-02: 1.2.0: - TimeCorrelator construction method can
//								 be chosen via argument
//		UL, 2001-03-30: 1.3.0: - added
//									TimeCorrelator::toOnBoardTime()
//							   - give some routines better names
//		UL, 2002-01-04: 1.4.0: - added
//									TimeCorrelator::constructionMethod()
//		UL, 2003-02-07: 1.5.0: - various changed related to new piecewise-
//								 linear interpolation of TC data (see doc
//								 for details)
//
//	USED ACRONYMS:
//		AHF	: Attitude History File
//		ERT	: Earth Received Time
//		OBT	: On-Board-Time
//		ODF	: Obervation Data File
//		TCF	: Time Correlation File
//		RMS	: Root-Mean-Square
//		UTC	: Coordinated Universal Time
//		MET : Mission-Ellapsed-Time (number of elapsed seconds since
//		      fixed mission reference time 1998-01-01T00:00:00 TT)
//
//	AUTHOR:
//		UL, Wed Jul 15 14:38:04 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: ICD: Observation and Slew Data Files, XMM-SOC-ICD-0004-SSD,
//			 Issue 1.0, K. Galloway
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/oal/src/TimeCorrelator.h,v 1.14 2003/06/16 08:41:35 gvacanti Exp $
///

#ifndef TIME_CORRELATOR_H
#define TIME_CORRELATOR_H

#include "Odf.h"
#include "Xmm.h"
#include "STime.h"
#include "OalChild.h"

#include <vector>

class OalServer;
class TimeCorrelationDataInterpolator;	// forward declaration

class TimeCorrelator : public virtual OalChild<OalServer> {
public:
	TimeCorrelator(const STime &timeBase = Xmm::missionReferenceTime);
		//
		//	PURPOSE:
		//		Create dummy instance of TimeCorrelator object
		//
		//	ARGUMENTS:
		//		I:	timeBase
		//				the time base all to be returned time tags will refer
		//				to
		//

	typedef struct {
		short int	degree;			//	degree of polynomial; neagtive values
									//	mean piecewise-linear
		double		rms;			//	RMS - quality of fit
		double		discardPoints;	//	fraction of points to be discared
									//	from data set to achive given RMS
		bool		ignoreGSHandovers;
									//	if set to yes, time correlation
									//	data after a GS handover are ignored
	} DataInterpolatorSpecs;

	enum ConstructionMethod { TCS_ONLY = 0, TCS_RECALCULATE_DELAYS };
	TimeCorrelator(const OalServer *oalEngine,
				   const DataInterpolatorSpecs &obtMetDataInterpolatorSpecs,
				   const DataInterpolatorSpecs &ertMetDataInterpolatorSpecs,
				   ConstructionMethod cm = TCS_ONLY,
				   const STime &timeBase = Xmm::missionReferenceTime);

	TimeCorrelator(const OdfSet *timeCorrelationSet,
				   const DataInterpolatorSpecs &obtMetDataInterpolatorSpecs,
				   const DataInterpolatorSpecs &ertMetDataInterpolatorSpecs,
				   ConstructionMethod cm = TCS_ONLY,
				   const STime &timeBase = Xmm::missionReferenceTime);

	TimeCorrelator(const std::string &timeCorrelationSetName,
				   const DataInterpolatorSpecs &obtMetDataInterpolatorSpecs,
				   const DataInterpolatorSpecs &ertMetDataInterpolatorSpecs,
				   ConstructionMethod cm = TCS_ONLY,
				   const STime &timeBase = Xmm::missionReferenceTime);
		//
		//	PURPOSE:
		//		Create instance of TimeCorrelator object from contents
		//		of ODF TCF; the latter contains the OBT-to-UTC relation in
		//		a tabulated form, i.e. a list of diskrete OBT-UTC data points.
		//		This data will be read and fitted to to a polynomal of a
		//		maximum allowed degree which will be used by the
		//		below methods toMissionEllapsedTime() to perform
		//		conversions from the OBT/ERT to MET. Leap seconds in UTC are
		//		properly treated.
		//
		//	ARGUMENTS:
		//		I:	timeCorrelationSet
		//				must refer to an ODF TCF otherwise an error condition
		//				is raised
		//
		//			timeCorrelationSetName
		//				the full name (dir+basename) of an ODF TCF; an
		//				error condition is raised if the specified file is not
		//				valid
		//
		//			timeBase
		//				the time base all to be returned time tags will refer
		//				to
		//			obtMetDataInterpolatorSpecs
		//				see description in constructor of
		//				TimeCorrelationDataInterpolator() below
		//


	ConstructionMethod constructionMethod() const
	{ return _constructionMethod; }
		//
		//	PURPOSE:
		//		Inquire method of OBT-MET construction method
		//


	const DataInterpolatorSpecs &obtMetDataInterpolatorSpecs() const
	{ return _obtMetDataInterpolatorSpecs; }
	const DataInterpolatorSpecs &ertMetDataInterpolatorSpecs() const
	{ return _ertMetDataInterpolatorSpecs; }
		//
		//	PURPOSE:
		//		Inquire data of the fit OBT/ERT-UTC fit polynomial;
		//
		//	RETURNS:
		//		structure with the following fields:
		//			<degree>		: actual degree of the fit polynomial
		//							  can be lower or equal to the value
		//							  specified in construtor call
		//			<rms>			: RMS as a measure of the quality
		//							  of the fit
		//			<discardPoints>	: fraction of points discarded from
		//							  original data set in ODF TCF
		//


	void setBaseTime(const STime &newBaseTime) const;
		//
		//	PURPOSE:
		//		Change the base time specified in any of the above
		//		constructors
		//
		//	ARGUMENTS:
		//		I:	newBaseTime
		//				the new base time to be used
		//


	double toMissionEllapsedTime(double obtTag) const;
	double toMissionEllapsedTime(unsigned obtCoarse,
								 unsigned short obtFine) const;
		//
		//	PURPOSE:
		//		Methods to convert a given On-Board-Time (OBT) tag into
		//		a offset time in seconds measured from the base time
		//		specified in the constructor calls or via
		//		TimeCorrelator::setBaseTime(). The OBT-UTC interpolation
		//		curve which has been set up during the object construction
		//		is employed. The quantity returned is directly usable for
		//		the time tag fields in event lists and HK files. The OBT
		//		values are supposed to monotonically increase from call to
		//		call. A decrease with respect to the previous value is
		//		interpreted as a clock wrap-around which is then properly
		//		handled.
		//
		//		Please note: TimeCorrelator::resetObtMetRelation
		//					 below resets the OBT-MET conversion, i.e.,
		//					 a decreasing OBT value is then _not_ considered
		//					 as a clock wrap around
		//
		//		ARGUMENTS:
		//			I:	obtCoarse/obtFine
		//					Coarse and fine component of an OBT tag
		//					as a pair of 32 and 16 bit unsigned values.
		//					The coarse time is a integral second
		//					counter, while the fine time measures the
		//					fractional part of a second with a resolution
		//					of 2^(-16) s = 1/65536 sec.
		//
		//				obttag
		//					A Coarse/Fine time pair combined as a double:
		//						obttag = obtCoarse + obtFine/65536.
		//
		//		RETURNS:
		//			The difference in seconds to the base time (mission
		//			reference time) corresponding to a given OBT value.
		//


	double toOnBoardTime(double missionEllapsedTime) const;
		//
		//	PURPOSE:
		//		The inverse of toMissionEllapsedTime(), i.e., convert
		//		a given MET to an OBT value
		//
		//	ARGUMENTS:
		//		missionEllapsedTime	: MET value [s]
		//
		//	RETURNS:
		//		OBT value corresponding to given MET
		//


	void resetObtMetRelation() const;
		//
		//	PURPOSE:
		//		Reset the OBT-MET relation. A call to this method
		//		between two calls TimeCorrelator::toMissionEllapsedTime() with
		//		a decreasing OBT value will prevent the decrease to be
		//		interpreted as a clock wrap around.
		//


	void plotObtMetRelation() const;
		//
		//	PURPOSE:
		//		Plot the OBT-MET relation using PGPLOT. The
		//		environment variable PGPLOT_DEV determines the
		//		plotting device.
		//


	double toMissionEllapsedTime(const std::string &ertString) const;
	double toMissionEllapsedTime(const STime &ertTag) const;
		//
		//	PURPOSE:
		//		Same as TimeCorrelator::toMissionEllapsedTime() above
		//		but using an ERT tag instead of OBT. Please note: ERT values
		//		cannot decrease - there is no wrap-around case.
		//
		//	ARGUMENTS:
		//		I:	ertString
		//				ERT tag in the format yyyy-mm-ddThh:mm:ss[.sss]
		//
		//			ertTag
		//				ERT tag as an STime object
		//
		//	RETURNS:
		//		The difference in seconds to the base time (mission
		//		reference time) corresponding to a given ERT value.
		//


	void plotErtMetRelation() const;
		//
		//	PURPOSE:
		//		Plot the ERT-MET relation using PGPLOT. The
		//		environment variable PGPLOT_DEV determines the
		//		plotting device.
		//


	const TimeCorrelationDataInterpolator *obt2MetInterpolator() const
	{ return _obtMetInterpolator; }
	const TimeCorrelationDataInterpolator *ert2MetInterpolator() const
	{ return _ertMetInterpolator; }
	const TimeCorrelationDataInterpolator *met2ObtInterpolator() const
	{ return _metObtInterpolator; }
		//
		//	PURPOSE:
		//		Give direct access to the OBT/ERTtoMET and METtoOBT
		//		interpolators
		//


	const TimeCorrelator &addCommonAttributes(Attributable *attrib) const;
		//	PURPOSE:
		//		Add the following list of attributes to passed attributable:
		//			TFIT_RPD =     T / signal propagation delays recalculated?
		//			TFIT_DEG =     1 / degree of OBT-MET fit polynomial
		//			TFIT_RMS = 1.e-2 / RMS of OBT-MET fit
		//			TFIT_PFR =   .02 / fraction of removed TCS data points
		//			TFIT_IGH =     T / Ignored GS handover in TCS data?
		//		If attributes exist values shall be replaced.
		//		


	~TimeCorrelator();
		//
		//	PURPOSE:
		//		Destroy TimeCorrelator object constructed with any of the
		//		above constructors
		//

private:
	//	methods
	void construct(const std::string &, bool recalcDelays = false);
	double metWithPropDelays(const std::string &ertraw,
							 const std::string &gsid);
	const TimeCorrelationDataInterpolator *
			constructInterpolator(DataInterpolatorSpecs &,
								  std::vector<double> &,
								  std::vector<double> &) const;

	//	data members
	DataInterpolatorSpecs	_obtMetDataInterpolatorSpecs,
							_ertMetDataInterpolatorSpecs;
	ConstructionMethod		_constructionMethod;
	mutable STime			_baseTime;
	STime					_utcBaseTime;
	double					_utcOffset;
	mutable double			_lastObtTag;
	mutable unsigned		_obtWraps;
	const TimeCorrelationDataInterpolator
							*_obtMetInterpolator,
							*_ertMetInterpolator,
							*_metObtInterpolator;
};


class TimeCorrelationDataInterpolator {
public:
	TimeCorrelationDataInterpolator() {}

	TimeCorrelationDataInterpolator(const std::vector<double> &x,
									const std::vector<double> &y,
									short int maxDegree = 1,
									double rms = 1.e-10) :
		_degree(maxDegree), _rms(rms), _x(x), _y(y) {}
		//
		//	PURPOSE:
		//		Construct from a given set of discrete data points a function
		//		object that can interpolate those data. Two distinct options
		//		chosen by the sign of <maxDegree> are supported:
		//			1. >=0: The interpolator is a polynomial of
		//					maximum degree <maxDegree> constructed with
		//					SLATEC-routine `DPOLFT'. This method automatically
		//					optimizes the degree of the polynomial based on
		//					a given RMS to achieved. If no satisfactory fit
		//					can be attained, an error condition will be raised
		//			2. <0 : The function object constructed does a piecewise-
		//					linear interpolation of the input data; i.e.
		//
		//	ARGUMENTS:
		//		I:	x
		//				abscissa values ...
		//			y
		//				... and corresponding ordinates
		//			maxDegree/rms
		//				specifications controlling the polynomial fit of the
		//				data; <maxDegree>
		//				defines the maximum allowed degree (e.g. 1 == linear)
		//				of the fit polynomal; <rms> is the to be obtained RMS
		//				value; it is possible that no polynomial with a degree
		//				lower than or equal to <polyDegree> can be found which
		//				satisfies the <rms> criterion. In this case it is
		//				attempted to achive the desired result by successively
		//				removing individial points from the data
		//				set: The respective point which is most "off" the
		//				computed fit polynomal will be removed first, the fit
		//				repeated, etc. till the requested RMS has been
		//				achieved or the fraction of removed points exceeds
		//				<discardPoints>. In the latter case a fatal error
		//				condition will be raised.
		//
		//				For negative maxDegree values the object will do
		//				a piecewise-linear interpolation of the input
		//				data <x>,<y>, such that evaluate(x[n])==y[n]
		//				(see below) shall be true for all 0<n<x.size()-1
		//


	virtual double evaluate(double x, bool warningOnExtrapolation = true)
	const = 0;
	virtual double derivative(double x, unsigned short n=1,
					  bool warningOnExtrapolation = true) const = 0;
		//
		//	PURPOSE:
		//		Evaluate fit polynomial/derivative of fit polynomial at a
		//		specific point
		//
		//	ARGUMENTS:
		//		value	: point at which polynomial is to be evaluated
		//		n		: derivative to evaluate (>=1)
		//		warningOnExtrapolation: Issue a warning if polynomial is
		//				  evaluated outside abscissa data range?
		//
		//	RETURNS:
		//		value of polynomial/derivative at <x>
		//


	short int degree() const { return _degree; }
	double rms() const { return _rms; }
		//
		//	PURPOSE:
		//		Return actual degree and achieved RMS of the fit polynomal;
		//		the values can deviate from the specifications given
		//		in the constructor call
		//


	const std::vector<double> &xPoints() const { return _x; }
	const std::vector<double> &yPoints() const { return _y; }
		//
		//	PURPOSE:
		//		Return input data points with which the object was constructed
		//


	virtual ~TimeCorrelationDataInterpolator() {}
		//
		//	PURPOSE:
		//		destroy object constructed with any of the above constructors
		//

protected:
	short int	_degree;	//	degree of fit polynomial
	double		_rms;		//	RMS - quality of fit
	std::vector<double>	_x, _y;	// input data points
};

std::ostream &operator<<(std::ostream &os,
						 const TimeCorrelationDataInterpolator &fp);

#endif
